/************************************************************************
*									*
*	DMX Dongle utility  V1.2  		  			*
*									*
*	(C) Copyright Artistic Licence (UK) Ltd 1995-96			*
*                                                                       *
*	NOT FOR DISTRIBUTION WITHOUT WRITTEN PERMISSION			*
*									*
*       Author W D Howell						*
*									*
*-----------------------------------------------------------------------*
*									*
*       ported to Linux by Michael Stickel (michael@cubic.org)		*
*									*
************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/io.h>
#include <unistd.h>

#include "dongle.h"

/* Raw data header for transmission to dongle 
   Direct access by library routines only  */

typedef struct 
{
  unsigned char	break_time;	/* DMX Break Time uS */
  unsigned char	mab_time;	/* DMX Mark after Break time uS */
  unsigned char	control_3;	/* Not used */
  unsigned char	rx_header;	/* DMX receive header code */
  unsigned char	rx_base_hi;	/* DMX receive base address */
  unsigned char	rx_base_lo;	/*  (range 1 - 512) */
  unsigned char	tx_header;	/* DMX transmit header code */
  unsigned char	aux1;		/* Not used */
  unsigned char	aux2;
  unsigned char	aux3;
  unsigned char	aux4;
  unsigned char	aux5;
  unsigned char	aux6;
  unsigned char	aux7;
  unsigned char	aux8;
  unsigned char	aux9;
  unsigned char	aux10;
} HEADER;

static HEADER header = {
  200,		/* break time */
  26,		/* mab time */
  0,		/* not used */
  0x00,		/* rx header */
  0x00,		/* rx base hi */
  0x01,		/* rx base lo */
  0x00,		/* tx header */
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0
};

/* Raw data returned by to dongle 
   Direct access by library routines only. 
   Applications should use RX_CTRL  */

typedef struct	
{
  unsigned char	check;	/* bit 7 = tx block, lo nib= dongle type */
  unsigned char	firmware;	/* firmware revision */
  unsigned char	chan_count_hi;	
  unsigned char	chan_count_lo;
  unsigned char	period_hi;
  unsigned char	period_lo;
  unsigned char	break_hi;
  unsigned char	break_lo;
  unsigned char	mab_hi;
  unsigned char	mab_lo;
  unsigned char	rx_break_count_hi;
  unsigned char	rx_break_count_lo;
  unsigned char	rx_ovr_count_hi;
  unsigned char rx_ovr_count_lo;
  unsigned char	rx_hdr_count_hi;
  unsigned char	rx_hdr_count_lo;
  unsigned char rx_frm_count_hi;
  unsigned char	rx_frm_count_lo;
} RET_DATA;



static RET_DATA ret_data = {0,0};	/* clear dongle type and firmware version */
static MEM      mem[MAX_MEM];
RX_CTRL  rx_ctrl;
RX       rx;



/* Define low level controls for dongle access */

#define strobe_hi       	0x04
#define	strobe_lo		0x05
#define d8		       	0x02

#define command_null		0x00
#define command_set_tx		0x02
#define command_set_rx		0x04
#define command_start_tx	0x06
#define command_break_tx	0x08
#define command_clear_buffer	0x0a
#define command_max_min		0x0c
#define command_compress	0x0d
#define command_decompress	0x0e

/* Global variables are all available through function calls */

static unsigned int address=0x378;	/* address value */
static char dongle_fail=NO;
static char dongle_compress=NO;	/* YES for rx packed two bit data */

static unsigned char dir=0xf0;		/* default to o/p for he */
/* Define low level calls. These are normally used by library
   and not called by application software. */


char tx_nibble       (unsigned int address, unsigned char data);
char tx_nibble_ctrl  (unsigned int address, unsigned char data);
char tx_control_block     (char);
char is_dmx_tx_block_finished  (void);
char power_on        (unsigned int address);
char power_off       (unsigned int address);
void clr_error_flag  (void);
void set_error_flag  (void);


/***********************************************
Name:	tx_byte_ctrl
Purp:	Send byte to dongle & flag control mode
Ret:	Next byte in sequence from dongle
***********************************************/
inline char tx_byte_ctrl(unsigned int address, unsigned char data)
{
  return (tx_nibble_ctrl(address, data>>4) & 0xf0) |
    ((tx_nibble(address, data) >> 4) & 0xf);
}

/***********************************************
Name:	tx_byte
Purp:	Send byte to dongle
Ret:	Next byte in sequence from dongle
***********************************************/
inline char tx_byte(unsigned int address, unsigned char data)
{
  return (tx_nibble(address, data>>4)&0xf0) |
    ((tx_nibble(address, data)>>4)&0xf) ;
}

inline char process_return_data(char ret) 
{ 
  return (ret<<1) & 0xf0; 
}


/***********************************************
Name:	init_dongle
Purp:	Start up dongle at specified address
Ret:	Returns low nibble of check code.
	==0x05 for standard dongle
***********************************************/
unsigned char init_dongle(unsigned int addr)
{
  if(addr==0)
    address=0x378;
  else
    address=addr;

  printf("init dongle ."); fflush(stdout);

  //clr_error_flag();
  power_off(address);
  sleep(1);

  printf("."); fflush(stdout);
  power_on(address);
  sleep(1);

  tx_control_block(command_null);
  return((ret_data.check)&0x0f);
}

/***********************************************
Name:	end_dongle
Purp:	Close down dongle. Final quit call 
Ret:	YES
***********************************************/
char end_dongle(void)
{
  power_off(address);
  return(YES);
}

/***********************************************
Name:	tx_dmx
Purp:	Transmit a 512 channel dmx frame using
	data from the mem array.
	Control holds in this routine until
	entire frame is transmitted
Ret:	YES
***********************************************/
char tx_dmx(memory)
{
  int   channel=0;

  if (get_firmware()>16)	/* can use 256 block size*/
    {
      header.aux1=0xff;
      tx_control_block(command_set_tx);
      while(channel<256)
	tx_byte(address,mem[memory].level[channel++]);
      
      tx_control_block(command_break_tx);
      tx_control_block(command_start_tx);
      is_dmx_tx_block_finished();
      
      tx_control_block(command_null);
      while(channel<512)
	tx_byte(address,mem[memory].level[channel++]);
      tx_control_block(command_start_tx);
      is_dmx_tx_block_finished();
      /* wait for frame to finish
	 so next call does not abort
	 current frame */

    }
  else		/* use 128 block size */
    {
      tx_control_block(command_set_tx);
      while(channel<128)
	tx_byte(address,mem[memory].level[channel++]);
      
      tx_control_block(command_break_tx);
      tx_control_block(command_start_tx);
      is_dmx_tx_block_finished();

      tx_control_block(command_null);
      while(channel<256)
	tx_byte(address,mem[memory].level[channel++]);
      tx_control_block(command_start_tx);
      is_dmx_tx_block_finished();

      tx_control_block(command_null);
      while(channel<384)
	tx_byte(address,mem[memory].level[channel++]);
      tx_control_block(command_start_tx);
      is_dmx_tx_block_finished();

      tx_control_block(command_null);
      while(channel<512)
	tx_byte(address,mem[memory].level[channel++]);
      tx_control_block(command_start_tx);
      is_dmx_tx_block_finished();
      /* wait for frame to finish
	 so next call does not abort
	 current frame */
    }
  return(YES);
}


/************************************
   name:     tx_control_block
   function: Transmit header block to dongle (HEADER)
             Receive header block from dongle (RET_DATA)
             Update RX_CTRL structure
   return:   YES
************************************/
char tx_control_block(char command)
{
  if(command==command_set_tx)
    dir=0xf0;
  else if(command==command_set_rx)
    dir=0x70;

  ret_data.check=tx_byte_ctrl(address,command);

  ret_data.firmware          = tx_byte(address, header.break_time);
  ret_data.chan_count_hi     = tx_byte(address, header.mab_time);
  ret_data.chan_count_lo     = tx_byte(address, header.control_3);
  ret_data.period_hi         = tx_byte(address, header.rx_header);
  ret_data.period_lo         = tx_byte(address, header.rx_base_hi);
  ret_data.break_hi          = tx_byte(address, header.rx_base_lo);
  ret_data.break_lo          = tx_byte(address, header.tx_header);
  ret_data.mab_hi            = tx_byte(address, header.aux1);
  ret_data.mab_lo            = tx_byte(address, header.aux2);
  ret_data.rx_break_count_hi = tx_byte(address, header.aux3);
  ret_data.rx_break_count_lo = tx_byte(address, header.aux4);
  ret_data.rx_ovr_count_hi   = tx_byte(address, header.aux5);
  ret_data.rx_ovr_count_lo   = tx_byte(address, header.aux6);
  ret_data.rx_hdr_count_hi   = tx_byte(address, header.aux7);
  ret_data.rx_hdr_count_lo   = tx_byte(address, header.aux8);
  ret_data.rx_frm_count_hi   = tx_byte(address, header.aux9);
  ret_data.rx_frm_count_lo   = tx_byte(address, header.aux10);

  /* convert raw return data to RX_CTRL */

  rx_ctrl.channel_count = ((ret_data.chan_count_hi)<<8) 
                        | ret_data.chan_count_lo;
  if (rx_ctrl.channel_count > 511) rx_ctrl.channel_count = 512;
  rx_ctrl.period        = ((ret_data.period_hi)<<8)
                        | ret_data.period_lo;
  rx_ctrl.cap1          = ((ret_data.break_hi)<<8) | ret_data.break_lo;
  rx_ctrl.cap2          = ((ret_data.mab_hi)<<8)   | ret_data.mab_lo;
  rx_ctrl.brk           = (rx_ctrl.cap2-rx_ctrl.cap1)/4;
  rx_ctrl.mab           = rx_ctrl.cap1/4+48;

  if(rx_ctrl.period==0) rx_ctrl.period=1;

  rx_ctrl.frequency=1000/rx_ctrl.period;
  rx_ctrl.break_count = ((((ret_data.rx_break_count_hi)<<8)
                      | ret_data.rx_break_count_lo)&0xffff);
  rx_ctrl.hdr_count = ((((ret_data.rx_hdr_count_hi)<<8)
                      | ret_data.rx_hdr_count_lo)&0xffff);
  rx_ctrl.frm_count = ((((ret_data.rx_frm_count_hi)<<8)
                      | ret_data.rx_frm_count_lo)&0xffff);
  rx_ctrl.ovr_count = ((((ret_data.rx_ovr_count_hi)<<8)
                      | ret_data.rx_ovr_count_lo)&0xffff);
  return(YES);
}

/*
name:	is_dmx_block_finished
Purp:	Hold until dongle has finished a 128
	byte block transmission
Ret:	YES
***********************************************/
char is_dmx_tx_block_finished(void)
{
  //if(is_error_flag())
  //  return (NO);
  while ( (!tx_byte_ctrl(address,0x00)) &0x80);
  return(YES);
}


/***********************************************
Name:	tx_nibble
Purp:	Send nibble to dongle
Ret:	Next nibble in sequence from dongle
	0 if error flag set by timeout
***********************************************/
char tx_nibble(unsigned int address, unsigned char data)
{
  long  timeout=TIMEOUT;
  char  ret;

  outb(strobe_hi+d8, address+2);   /* strb hi */
  outb((data&0xf) | dir, address); /* data out*/
  outb(strobe_lo+d8, address+2);   /* strb lo */

  while(!((ret=inb(address+1))&0x80))
    {
      if(timeout--==0)
	{
	  //set_error_flag();
	  return(0);
	}
    }
  
  outb(strobe_hi+d8, address+2);		/* strb hi */

#if 1
  timeout=TIMEOUT;
  while( (inb(address+1)&0x80))
    {
#if 1
      if(timeout--==0)
	{
	  //set_error_flag();
	  return(0);
	}
#endif
    }
#endif

  //clr_error_flag();
  return (process_return_data(ret));
}



/***********************************************
Name:	tx_nibble_ctrl
Purp:	Send nibble to dongle & flag control mode
Ret:	Next nibble in sequence from dongle
	0 if error flag set by timeout
***********************************************/
char tx_nibble_ctrl(unsigned int address, unsigned char data)
{
  long timeout=10000;
  char ret;

  outb(strobe_hi,        address+2); /* strb hi */
  outb((data&0xf) | dir, address);   /* data out*/
  outb(strobe_lo,        address+2); /* strb lo */
  
  while(!(ret=(inb(address+1))&0x80))
    {
      if(timeout--==0)
	{
	  //set_error_flag();
	  return(0);
	}
      usleep(100);
    }

  outb(strobe_hi, address+2);		/* strb hi */

  timeout=10000;
  while( (inb(address+1)&0x80))
    {
      if(timeout--==0)
	{
	  //set_error_flag();
	  return(0);
	}
      usleep(100);
    }

  //clr_error_flag();
  return(process_return_data(ret));
}



/***********************************************
Name:	rx_dmx
Purp:	Update RX structure with 128 bytes of
	DMX data (starting at base address).
	Max & Min values also updated.
Note:	Designed to operate with dongle in 
	uncompressed mode

Ret:	YES
***********************************************/
char rx_dmx(void)
{
  int           channel=0;
  int           max=128;
  unsigned char level;
  unsigned int  base;

  if(get_firmware()>16)	/* can use 256 block size*/
    max=256;
  else
    max=128;

  base=((header.rx_base_hi<<8)&0xff00) + header.rx_base_lo - 1;

  tx_control_block(command_set_rx);
  for (channel=0; channel<max; channel++)
    {
      level=tx_byte(address,0x80);
      rx.cur_level[channel+base]=level;
      if(rx.max_level[channel+base]<level)
	rx.max_level[channel+base]=level;
      if(rx.min_level[channel+base]>level)
	rx.min_level[channel+base]=level;
    }
  return(YES);
}


#if 0
/***********************************************
Name:	wait_for_break
Purp:	Hold until dongle receives a break
Ret:	YES when break detected
	NO if timeout
***********************************************/
char wait_for_break(void)
{
  unsigned int last_count;
  long timeout=100;
  
  last_count=rx_ctrl.break_count;
  while(last_count==rx_ctrl.break_count)
    {
      tx_control_block(command_null);
      if(timeout--==0)
	return(NO);
      usleep(10);
    }
  return(YES);
}
#endif

#if 0
/***********************************************
Name:	rx_reset_counters
Purp:	Reset dongle error and frame counters
	to zero
Ret:	YES
***********************************************/
char rx_reset_counters(void)
{
  tx_control_block(command_max_min);
  return(YES);
}
#endif

#if 1
/***********************************************
Name:	reset_rx_base
Purp:	Set new DMX receive bas address
	range 0 to 511
Ret:	YES
***********************************************/
char reset_rx_base(unsigned int base)
{
  base++;
  if(base>512)
    base=512;
  header.rx_base_hi=(base>>8)&0xff;
  header.rx_base_lo=base&0xff;
  
  tx_control_block(command_null);
  return(YES);
}
#endif

/***********************************************
Name:	power_on
Purp:	Dongle out of reset
Ret:	YES
***********************************************/
char power_on(unsigned int address)
{
  outb(strobe_hi, address+2);
  outb(0xff,address);
  return(YES);
}

/***********************************************
Name:	power_off
Purp:	Dongle reset
Ret:	YES
***********************************************/
char power_off(unsigned int address)
{
  outb(0x80, address+2);
  outb(0x00, address);
  return(YES);
}

#if 0
/***********************************************
Name:	clr_error_flag
Purp:	Show good Dongle communication
Ret:	Void
***********************************************/
void clr_error_flag(void)
{
  dongle_fail=NO;
}

/***********************************************
Name:	set_error_flag
Purp:	Show Dongle communication fail
Ret:	Void
***********************************************/
void set_error_flag(void)
{
  dongle_fail=YES;
}

/***********************************************
Name:	is_error_flag
Purp:	Return Dongle communication status
Ret:	YES/NO
***********************************************/
char is_error_flag(void)
{
  return(dongle_fail);
}
#endif

#if 0
/***********************************************
Name:	clr_memory
Purp:	Zero entire mem array
Ret:	Void
***********************************************/
void clr_memory(void)
{
  int memory, channel;
  for(memory=0; memory<MAX_MEM; memory++)
    for(channel=0; channel<MAX_CHAN; channel++)
      mem[memory].level[channel]=0;

}
#endif

#if 0
/***********************************************
Name:	clr_rx
Purp:	Clear receive data buffer
Ret:	Void
***********************************************/
void clr_rx(void)
{
  int channel;
  for(channel=0; channel<MAX_CHAN; channel++)
    {
      rx.cur_level[channel]=0;
      rx.min_level[channel]=0;
      rx.max_level[channel]=0;
    }
}
#endif

#if 1
/***********************************************
Name:	set_rx_header
Purp:	Set new receive DMX header code
Ret:	header code
***********************************************/
unsigned char set_rx_header(unsigned char rheader)
{
  header.rx_header=rheader;
  return(rheader);
}
#endif

#if 0
/***********************************************
Name:	get_rx_header
Purp:	Return current receive DMX header code
Ret:	header code
***********************************************/
unsigned char get_rx_header(void)
{
  return(header.rx_header);
}
#endif

#if 0
/***********************************************
Name:	set_tx_header
Purp:	Set new transmit DMX header code
Ret:	header code
***********************************************/
unsigned char set_tx_header(unsigned char theader)
{
  header.tx_header=theader;
  return(theader);
}
#endif

#if 0
/***********************************************
Name:	get_tx_header
Purp:	Return current transmit DMX header code
Ret:	header code
***********************************************/
unsigned char get_tx_header(void)
{
  return(header.tx_header);
}
#endif

#if 0
/***********************************************
Name:	set_tx_break_time
Purp:	Set new transmit break time (units = uS)
Ret:	break time
***********************************************/
unsigned char set_tx_break_time(unsigned char break_time)
{
  header.break_time=break_time;
  return(break_time);
}
#endif

#if 0
/***********************************************
Name:	get_tx_break_time
Purp:	Get current transmit break time (units = uS)
Ret:	break time
***********************************************/
unsigned char get_tx_break_time(void)
{
  return(header.break_time);
}
#endif

#if 0
/***********************************************
Name:	set_tx_mab_time
Purp:	Set new transmit mark after break time 
	Units = uS
	Available values 2,6,10,14,18,22,26..
Ret:	MaB time
***********************************************/
unsigned char set_tx_mab_time(unsigned char mab_time)
{
  header.mab_time=mab_time;
  return(mab_time);
}
#endif

#if 0
/***********************************************
Name:	get_tx_mab_time
Purp:	Get current transmit mark after break time 
	Units = uS
Ret:	MaB time
***********************************************/
unsigned char get_tx_mab_time(void)
{
  return(header.mab_time);
}
#endif

#if 0
/***********************************************
Name:	set_tx_aux1
Purp:	Set new transmit block size
Ret:	block size
***********************************************/
unsigned char set_tx_aux1(unsigned char aux1)
{
  header.aux1=aux1;
  return(aux1);
}
#endif

#if 0
/***********************************************
Name:	get_tx_aux1
Purp:	Get current block size ( number bytes dmx transmiited
	each time a command_tx_start is issued)
Ret:	block size
***********************************************/
unsigned char get_tx_aux1(void)
{
  return(header.aux1);
}
#endif

#if 1
/***********************************************
Name:	set_compress
Purp:	Set Dongle compress mode.
	YES = Receives 512 channels @ 2 bit 
	resolution compressed into 128 bytes
	NO = Receives 128 bytes @ 8 bit
	resolution
Ret:	YES/NO
***********************************************/
unsigned char set_compress(unsigned char compress)
{
  dongle_compress=compress;
  if(dongle_compress==YES)
    tx_control_block(command_compress);
  else
    tx_control_block(command_decompress);

  return(dongle_compress);
}
#endif

/***********************************************
Name:	get_compress
Purp:	Get Dongle compress mode.
	YES = Receives 512 channels @ 2 bit 
	resolution compressed into 128 butes
	NO = Receives 128 bytes @ 8 bit
	resolution
Ret:	YES/NO
***********************************************/
unsigned char get_compress(void)
{
  return(dongle_compress);
}

/***********************************************
Name:	get_firmware
Purp:	Return Dongle firmware version
Ret:	Firmware version
***********************************************/
unsigned char get_firmware(void)
{
  tx_control_block(command_null);
  return(ret_data.firmware);
}

/***********************************************
Name:	get_dongle_type
Purp:	Return Dongle Type
Ret:	0x05
***********************************************/
unsigned char get_dongle_type(void)
{
  tx_control_block(command_null);
  return((ret_data.check)&0x0f);
}
